#!/usr/bin/perl
#require 5.008;
our $VERSION = "0.04"; #Time-stamp: <2011-04-04T11:32:55Z>

##
## License:
##
##   Public Domain as a sort of mathematical formulae.
##
## Author's Link:
##
##   http://jrf.aboutme.jp/user_statuses/show/137725
##   http://jrf.cocolog-nifty.com/
##   (The pages are written in Japanese.)
##

use strict;
use warnings;
use utf8; # Japanese
use POSIX;


my $R = 0.15; # 最高利率から計算された BI 喪失時貸付利率。
my $P1 = 160; # 毎期、債務者から元債権者に支払われる額。
my $X1_0 = 1000; # 元債権者から債務者への貸付総額初期値。
my $X2_0 = 100; # 債務者から新債権者に送られたとされるポイント総額初期値。

my $Q = ceil(calc_q($X1_0, $X2_0)); # 新債権者が債務者分として最初に供託積み立てした額。


# $x1cur -- 元債権者から債務者への貸付総額。
# $x2cur -- 債務者から新債権者に送られたとされるポイント総額。
# r3($x1cur) -- 債務者の利子外の実返済率。
# p2($x1cur, $x2cur) -- 毎期、新債権者が元債権者に支払われる額。

sub x1next {
  my ($x1cur) = @_;

  return $x1cur - ($P1 - $R * $x1cur);
}

sub r3 {
  my ($x1cur) = @_;
  return ($P1 - $R * $x1cur) / $x1cur;
}

sub x2next {
  my ($x1cur, $x2cur) = @_;

  return $x2cur * (1 - r3($x1cur));
}

sub p2 {
  my ($x1cur, $x2cur) = @_;

  return $R * $x2cur * r3($x1cur);
}

# 解析後の関数。
sub x1b {
  my ($n) = @_;
  my $r = (1 + $R);
  my $k = - $P1;
  my $k1 = - $k / (1 - $r);
#  return ($r ** $n) * ($X1_0 + $k1) - $k1;
  return ((1 + $R) ** $n) * ($X1_0 - $Q - $P1/$R) + $P1/$R;
}

sub p2next {
  my ($x1cur, $x2cur, $p2cur) = @_;
  return $R * ($X2_0 / ($X1_0 - $Q)) * $P1 - $R * $R * $x2cur + $R * $p2cur;
}

sub q2 {
  my ($n) = @_;
  my $r = $R;
  my $k = $R * ($X2_0 / ($X1_0 - $Q)) * $P1;
  my $k1 = - $k / (1 - $r);
  my $a0 = p2($X1_0 - $Q, $X2_0);
  return ($r ** $n) * ($a0 + $k1) - $k1;
}

# sum_q2 は $P1 -> $R * $X1_0 で発散しうる。
# これを防ぐには、Sum P2 の中に元本返済に相当する分が少しでもあればいい。
# そこで $Q を導入した。

sub sum_q2 {
  my ($n) = @_;
  my $r = $R;
  my $k = $R * ($X2_0 / ($X1_0 - $Q)) * $P1;
  my $k1 = - $k / (1 - $r);
  my $a0 = p2($X1_0 - $Q, $X2_0);
#  print "a0 + k1 = " . ($a0 + $k1) . "\n";
#  print "- n * k1 = " . (- $n * $k1) . "\n";
  return (1 - ($r ** $n))/(1 - $r) * ($a0 + $k1) - $n * $k1;
}

sub nlast {
  return log($P1 / ($P1 - $R * ($X1_0 - $Q))) / log(1 + $R);
}

# 二分法(?)で X1(T) X2(T) に関し望ましい Q を求める。
sub calc_q {
  my ($x1_0, $x2_0) = @_;

  sub sum_q2b {
    my ($x1_0, $x2_0, $q) = @_;
    my $p1 = $R * $x1_0;
    my $tq = log($p1 / ($p1 - $R * ($x1_0 - $q))) / log(1 + $R);
    my $r = $R;
    my $k = $R * ($x2_0 / ($x1_0 - $q)) * $p1;
    my $k1 = - $k / (1 - $r);
    my $a0 = p2($x1_0 - $q, $x2_0);
    return (1 - ($r ** $tq))/(1 - $r) * ($a0 + $k1) - (1 + $tq) * $k1;
  }

  my $step = $x1_0 / 4;
  my $q = $x1_0 / 2;
  while (1) {
    my $qb = sum_q2b($x1_0, $x2_0, $q);
    if (abs($qb - $q) < 0.1) {
      last;
    }
    if ($qb < $q) {
      $q = $q - $step;
    } else {
      $q = $q + $step;
    }
    $step = $step / 2;
  }
  return $q;
}

MAIN:
{
  if ($R * $X1_0 > $P1) {
    print "No need of P2\n";
    exit(0);
  }

  print "init:\n\n";
  my $ret = 0;

  my $x1cur = $X1_0 - $Q;
  my $x2cur = $X2_0;
  my $n = 0;

  my $sum_q2 = 0;;
  my $qcur = ceil(calc_q($X1_0, $X2_0));
  my $sum_payback = 0;

  print "q = $qcur\n";

  print "\nloop:\n\n";

  while ($x1cur > 0 && $x2cur > 0) {
    print "x1 = $x1cur " . x1b($n) . "\n";
    print "x2 = $x2cur\n";
    if ($R * $x1cur > $P1) {
      print "No need of P2. Ended for uncertain condition.\n\n";
      $x1cur = x1next($x1cur);
      $x2cur = x2next($x1cur, $x2cur);
      $n++;
      last;
    } elsif ($x1cur < $P1) {
      print "No need of P2. Ended normally.\n\n";
      $x1cur -= $x1cur;
      $n++;
      next;
    }

    my $p2 = p2($x1cur, $x2cur);
    $ret += $p2;
    print "p2 =$p2\n";
    print "p2next = " . p2next($x1cur, $x2cur, $p2) . "\n";
    print "q2 = " . q2($n) . "\n";
    $sum_q2 += q2($n);

    print "x2/x1 = " . $x2cur / $x1cur . "\n";
    my $x1n = x1next($x1cur);
    my $x2n = x2next($x1cur, $x2cur);
    print "x2n/x2cur = " . $x2n / $x2cur . "\n";
    print "x1n/x1cur = " . $x1n / $x1cur . "\n";

    if ($x1n > 0 && $x2n > 0) {
      my $qn = calc_q($x1n + $qcur, $x2n);
      my $payback = floor($qcur - $qn - $p2);
      print "calc_q = " . $qn . "\n";
      print "payback = " . $payback . "\n";
      $qcur =  $qn - ($p2 - ($qcur - $qn - $payback));
      print "next qcur = " . $qcur . "\n";
      $sum_payback += $payback;
    } else {
      my $payback = $qcur - p2($x1cur, $x2cur);
      print "payback = " . $payback . "\n";
      $sum_payback += $payback;
      $qcur = 0;
    }

    $x1cur = $x1n;
    $x2cur = $x2n;
    $n++;
    print "\n";
  }
  if ($qcur) {
    my $payback = floor($qcur);
    print "last payback = " . $payback . "\n\n";
    $sum_payback += $payback;
  }

  print "result : \n\n";
  print "Sum p2 = $ret\n";
  print "Sum payback = $sum_payback\n";
  print "n = $n < " . nlast() . " + 1\n";
  print "sum_q2 = $sum_q2 " . sum_q2($n) . "\n";
}

